
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmhtml.scm
;; DESCRIPTION : conversion of TeXmacs trees into Html trees
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert tmhtml tmhtml)
  (:use (convert tools tmlength) (convert tools tmtable)
	(convert tools sxml) (convert tools stm) (convert tools sxhtml)
	(convert tmhtml htmlout))
  (:export tmhtml-env-patch tmhtml-produce tmhtml))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Postprocessing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-post-simplify-nodes l)
  ;; Catenate adjacent string nodes and remove empty string nodes
  (let rec ((l l))
    (cond ((null? l) '())
	  ((and (string? (car l)) (string-null? (car l)))
	   (rec (cdr l)))
	  ((null? (cdr l)) l)
	  ((and (string? (car l)) (string? (cadr l)))
	   (rec (cons (string-append (car l) (cadr l)) (cddr l))))
	  (else (cons (car l) (rec (cdr l)))))))

(define (tmhtml-post-simplify-element e)
  ;; Simplify the nodes of the element content
  (list (append (sxml-element-head e)
		(tmhtml-post-simplify-nodes (sxml-content e)))))

(define (tmhtml-post-heading l)
  ;; Post-process the converted result of a concat containing a section title.
  ;;
  ;; The first label after the section is changed to an 'id' attribute in the
  ;; heading element, if it has not already an 'id' attribute.
  ;;
  ;; NOTE: assumes the heading is the first node of the set.
  (let ((heading (car l)))
    (if (sxml-attr heading 'id) l
	(receive (labels rest) (list-partition (cdr l) sxhtml-label?)
	  (if (null? labels) l
	      (cons (sxml-prepend (sxhtml-glue-label heading (car labels))
				  (cdr labels))
		    rest))))))

(define (tmhtml-post-table l)
  ;; Post process the converted result of a concat containing a table.
  ;;
  ;; If a label is adjacent to the table, use it to set the table id. If there
  ;; are several labels adjacent to the table, leave all but one label
  ;; untouched. There is no guarantee on which label is glued.
  (define (glue-label-to-table x knil)
    (cond ((null? knil) (list x))
	  ((and (sxhtml-label? x)
		(sxhtml-table? (car knil))
		(not (sxml-attr (car knil) 'id)))
	   (cons (sxhtml-glue-label (car knil) x)
		 (cdr knil)))
	  ((and (sxhtml-table? x)
		(not (sxml-attr x 'id))
		(sxhtml-label? (car knil)))
	   (cons (sxhtml-glue-label x (car knil))
		 (cdr knil)))
	  (else (cons x knil))))
  (list-fold-right glue-label-to-table '() l))

(define (tmhtml-post-paragraphs l)
  ;; Post process a collection of h:p elements
  ;;
  ;; If a h:p contains a h:hN, remove the h:p node and prepend the rest of the
  ;; contents to the next h:p. If the next element, after post processing is
  ;; not a h:p, create an intermediate h:p to hold the data.
  ;;
  ;; If a h:p contains a list element, remove the enclosing h:p. The TeXmacs
  ;; editor ensures that an <item-list> or <desc-list> is the only element
  ;; contained in its enclosing <doc-item>.
  ;;
  ;; NOTE: assumes the heading is at the start of a paragraph. That is
  ;; consistent with the fact that (as of 2003-02-04) the only converted
  ;; invisible markup is <label> and correct usage requires it to be after the
  ;; section title.
  (let rec ((in l) (out '()) (trail #f))
    (let* ((para (and (pair? in) (car in)))
	   (cont (and para (sxml-content para)))
	   (first (and cont (pair? cont) (car cont)))
	   (next (lambda (o t) (rec (cdr in) o t)))
	   (flush (lambda () (if trail `((h:p ,@trail) ,@out) out)))
	   (accept (lambda () (if trail (sxml-prepend para trail) para)))
	   (give (lambda () (and (pair? (cdr cont)) (cdr cont)))))
      ;; invariant: (xor prev prev-trail)
      (cond ((null? in) (reverse (flush)))
	    ((or (null? cont) (string? first))
	     (next (cons (accept) out) #f))
	    ((sxhtml-heading? first)
	     ;; tmhtml-post-heading should be called by concat handler
	     (next (cons first (flush)) (give)))
	    ((sxhtml-list? first)
	     ;; texmacs editor ensures there is not trail after a list
	     (next (append cont (flush)) #f))
	    ((and (sxhtml-table? first) (null? (cdr cont)))
	     ;; if table is not alone, we cannot help but produce bad html
	     ;; if table is alone, drop the enclosing <h:p>
	     (next (append cont (flush)) #f))
	    (else (next (cons (accept) out) #f))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Strings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-string s) (cork->utf8 s))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Simple text
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-noop l) '())

(define (tmhtml-file l)
  (let* ((doc (car l))
	 (styles (cdadr l))
	 (lang (caddr l))
	 (tmpath (cadddr l)))
    `(h:html
      (h:head
       (h:title "no title")
       (h:meta (@ (name "generator")
		  (content ,(string-append "TeXmacs " (texmacs-version))))))
      (h:body (@ (align "justify"))
	      ,@(tmhtml doc)))))

(define (tmhtml-document l)
  (tmhtml-post-paragraphs (map (lambda (x) (cons 'h:p (tmhtml x)))
			       l)))

(define (tmhtml-paragraph l)
  (let rec ((l l))
    (if (null? l) '()
	(let ((h (tmhtml (car l)))
	      (r (rec (cdr l))))
	  (cond ((null? h) r)		; correct when r is null too
		((null? r) h)
		(else `(,@h (h:br) ,@r)))))))

(define (tmhtml-concat l)
  (tmhtml-post-simplify-nodes
   (let ((l (tmhtml-list l)))
     (cond ((null? l) '())
	   ((string? (car l)) l)
	   ((sxhtml-heading? (car l)) (tmhtml-post-heading l))
	   ((list-any sxhtml-table? l) (tmhtml-post-table l))
	   (else l)))))

(define (tmhtml-format l)
  (let ((s (car l)))
    (cond ((== s "new line") '((h:br)))
	  ((== s "next line") '((h:br)))
	  (else '()))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Local and global environment changes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-with-font-size val)
  (let* ((x (* (string->number val) 100))
	 (s (cond ((< x 1) #f) ((< x 55) "-4") ((< x 65) "-3")
		  ((< x 75) "-2") ((< x 95) "-1") ((< x 115) "0")
		  ((< x 135) "+1") ((< x 155) "+2") ((< x 185) "+3")
		  ((< x 225) "+4") ((< x 500) "+5") (else #f))))
    (and s `(h:font (@ (size ,s))))))

(define (tmhtml-with-one var val arg-result)
  ;; Internal routine for tmhtml-with.
  ;;
  ;; Convert a with-binding. Val is the value being bound to var. Arg-result is
  ;; the body of the with element (including subsequent bindings) after
  ;; conversion.
  (let ((w (or (drd-ref tmhtml-with-cmd% (list var val))
	       (let ((h (drd-ref tmhtml-with-cmd% var)))
		 (and h (h val))))))
    (if w
	(list (append w arg-result))
	arg-result)))

(define (tmhtml-with l)
  ;; Handler for 'with' primitive.
  ;;
  ;; Apply tmhtml-with-one for each binding in the node, and update the
  ;; environment. Earlier bindings will produce outer nodes.
  (let rec ((l l))
    ;; sanity check, only true if 'with' has an even number of
    ;; parameters, which should never happen.
    (if (null? l) '()
	(if (null? (cdr l))
	    (tmhtml (car l)) ;; all binding have been handled, convert body
	    ;; WARNING: bad conversion if variable or value is not a string.
	    (let* ((var (force-string (car l)))
		   (val (force-string (cadr l)))
		   (next (cddr l)))
	      (tmhtml-with-one var val (rec next)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other primitives
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-label l)
  ;; WARNING: bad conversion if ID is not a string.
  `((h:a (@ (id ,(force-string (car l)))))))

;(define (tmhtml-reference l)
;  (list 'ref (force-string (car l))))

;(define (tmhtml-pageref l)
;  (list 'pageref (force-string (car l))))

(define (tmhtml-hyperlink l)
  ;; WARNING: bad conversion if URI is not a string.
  ;; TODO: change label at start of content into ID attribute, move other
  ;; labels out (A elements cannot be nested!).
  `((h:a (@ (href ,(force-string (second l))))
	 ,@(tmhtml (first l)))))

(define (tmhtml-specific l)
  (if (== (car l) "html") (tmhtml (cadr l)) '()))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Standard markup
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-list-document list-doc)
  ;; Convert a list-document to a list of <h:li> elements.
  (define (item->li mark item)
    (cond ((null? item) '(h:li))
	  ((null? (cdr item)) `(h:li ,@(tmhtml (car item))))
	  (else `(h:li ,@(tmhtml `(document ,@item))))))
  (if (null? (cdr list-doc)) '((h:li))
      (stm-list-map item->li
		    (lambda (x) (== x '(item)))
		    (cdr list-doc))))

;; TODO: when the first data of the list is a label, it must be used to set the
;; ID attribute of the resulting xhtml element. When that is done, remove the
;; warning comment from htmltm-handler.

(define (tmhtml-itemize args)
  `((h:ul ,@(tmhtml-list-document (car args)))))

(define (tmhtml-enumerate args)
  `((h:ol ,@(tmhtml-list-document (car args)))))

(define (tmhtml-desc-document desc-doc)
  (define (item->dt-dd mark item)
    (let ((html-item (if (null? (cdr item))
			 (tmhtml (car item))
			 (tmhtml `(document ,@item)))))
      (append
       (if mark `((h:dt ,@(tmhtml (cadr mark)))) '())
       (cond ((and (null? html-item) mark) '())
	     ((null? html-item) '((h:dd)))
	     (else `((h:dd ,@html-item)))))))
  (if (null? (cdr desc-doc)) '((h:dd))
      (apply append (stm-list-map item->dt-dd
				  (lambda (x) (func? x 'item* 1))
				  (cdr desc-doc)))))

(define (tmhtml-description args)
  `((h:dl ,@(tmhtml-desc-document (car args)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Tables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-table-cols p)
  (map (lambda (x)
	 `(h:col (@ (align ,(cadr (assoc x '(("l" "left")
					     ("c" "center")
					     ("r" "right"))))))))
       (p 'cols 'halign)))

(define (tmhtml-table-contents p)
  (define (cell x)
    `(h:td ,@(tmhtml x)))
  (define (row l)
    `(h:tr ,@(map cell l)))
  (map row (p 'rows 'content)))

(define (tmhtml-table-make p)
  `((h:table
     ,@(if (p 'global 'border) '((@ (border "1"))) '())
     ,@(tmhtml-table-cols p)
     (h:tbody ,@(tmhtml-table-contents p)))))

(define (tmhtml-table halign border x)
  ;; assert (= 1 (length x))
  ;; assert (or (func? (car x) 'tformat) (func? (car x) 'table))
  (tmhtml-table-make
   (tmtable-parser `(tformat ,@(tmtable-cell-halign halign)
			     ,@(tmtable-block-borders border)
			     ,(car x)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pictures
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; WARNING: Should also test that width and height are not magnifications.
;; Currently, magnifications make string->tmlength return #f.

(define (tmhtml-dimension-attr s name handlers)
  (let ((w (and-let* ((tmlen (string->tmlength s))
		      ((not (tmlength-null? tmlen)))
		      (proc (assoc (tmlength-unit tmlen) handlers)))
		     ((second proc) (tmlength-value tmlen)))))
    (if w `((,name ,w)) '())))

(define (exact-number->string x)
  (number->string (inexact->exact x)))
(define (number->percent x)
  (string-append (exact-number->string (* 100 x)) "%"))

(define (tmhtml-postscript l)
  (let ((s (first l)) (w (second l)) (h (third l)))
    (if (not (string? s)) '()	; only convert linked images
	`((h:img (@ (src ,s)
		    ,@(tmhtml-dimension-attr
		       w 'width `((par ,number->percent)
				  (px ,exact-number->string)))
		    ,@(tmhtml-dimension-attr
		       h 'height `((px ,exact-number->string)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main conversion routines
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-list l)
  (append-map tmhtml l))

(define (tmhtml-dispatch htable l)
  (let ((x (drd-ref ,htable (car l))))
    (cond ((not x) #f)
	  ((procedure? x) (x (cdr l)))
	  (else (tmhtml-post-simplify-element
		 (append x (tmhtml-list (cdr l))))))))

(define (tmhtml-expand l)
  ;; Explicit expansions are converted and handled as implicit expansions.
  (tmhtml-implicit-expand (cons (string->symbol (car l))
				(cdr l))))

(define (tmhtml-implicit-expand l)
  (or (tmhtml-dispatch 'tmhtml-stdmarkup% l)
      (tmhtml-dispatch 'tmhtml-tables% l)
      '()))

(define (tmhtml x)
  ;; Main conversion function.
  ;; Takes a TeXmacs tree in Scheme notation and produce a SXML node-set.
  ;; All handler functions have a similar prototype.
  (if (string? x)
      (if (string-null? x) '()
	  (list (tmhtml-string x)))		; handle string nodes
      (or (tmhtml-dispatch 'tmhtml-primitives% x)
	  (tmhtml-implicit-expand x))))

(define (tmhtml-decorate root)
  "Add metadata to a XML document to prepare it for serialization"
  `(*TOP*
    ;; x-texmacs is internal encoding of TeXmacs.
    ;; It is Cork encoding with TeXmacs universal symbols.
    (*PI* xml "version=\"1.0\" encoding=\"x-texmacs\"")
    ,(sxml-set-attrs
      root '((xmlns:h "http://www.w3.org/1999/xhtml")
	     (xmlns:m "http://www.w3.org/1998/Math/MathML")
	     (xmlns:x "http://www.texmacs.org/2002/extensions")))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Dispatching
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(drd-dispatcher tmhtml-primitives%
  (document tmhtml-document)
  (para tmhtml-paragraph)
  (surround tmhtml-noop)
  (concat tmhtml-concat)
  (format tmhtml-format)
  ((:or hspace vspace* vspace space htab split move resize float
	repeat datoms dlines dpages dbox)
   tmhtml-noop)
  ((:or group left mid right big lprime rprime below above
	lsub lsup rsub rsup frac sqrt wide neg
	tree old_matrix old_table old_mosaic old_mosaic_item)
   tmhtml-noop)
  (table (lambda (l) (tmhtml `(tabular (table ,@l)))))
  (tformat (lambda (l) (tmhtml `(tabular (tformat ,@l)))))
  ((:or twith cwith tmarker row cell sub_table) tmhtml-noop)

  (assign tmhtml-noop)
  (with tmhtml-with)
  ((:or set reset) tmhtml-noop)
  ((:or var_expand expand apply) tmhtml-expand)
  ((:or begin end include macro func env eval) tmhtml-noop)
  (value tmhtml-expand)
  (arg tmhtml-noop)
  ((:or backup quote delay hold release) tmhtml-noop)
  
  ((:or or xor and not plus minus times over div mod merge length range
	number date translate is_tuple look_up equal unequal less lesseq
	greater greatereq if case while extern authorize)
   tmhtml-noop)
  ((:or inactive symbol latex hybrid tuple collection associate) tmhtml-noop)
  (label tmhtml-label)
  (reference tmhtml-noop)
  (pageref tmhtml-noop)
  (write tmhtml-noop)
  (specific tmhtml-specific)
  (hlink tmhtml-hyperlink)
  ((:or action tag meaning) tmhtml-noop)
  ((:or switch fold exclusive progressive superposed) tmhtml-noop)
  ((:or graphics point line arc bezier) tmhtml-noop)
  (postscript tmhtml-postscript)

  (!file tmhtml-file))

(drd-table tmhtml-stdmarkup%
  ;;; Sectioning
  (chapter (h:h1))
  (section (h:h2))
  (subsection (h:h3))
  (subsubsection (h:h4))
  ;; paragraph and subparagraph are intented to be used at the start
  ;; of a paragraph. So they cannot be converted to 'h5' and 'h6
  (paragraph (h:strong (@(class "paragraph"))))
  (subparagraph (h:strong (@(class "subparagraph"))))
  ;; Lists
  ((:or itemize itemize-minus itemize-dot itemize-arrow)
   ,tmhtml-itemize)
  ((:or enumerate enumerate-numeric enumerate-roman enumerate-Roman
	enumerate-alpha enumerate-Alpha)
   ,tmhtml-enumerate)
  ((:or description description-compact description-dash
	description-align description-long)
   ,tmhtml-description)
  ;; Phrase elements
  (strong (h:strong))
  (em (h:em))
  (dfn (h:dfn))
  (code* (h:code))
  (samp (h:samp)) ; WARNING: semantic documentation does not match HTML4
  (kbd (h:kbd))
  (var (h:var))
  (abbr (h:abbr))
  (acronym (h:acronym))
  ;; Presentation
  (tt (h:tt)))

;;    (name (h:name)) ; not in HTML4
;;    (person (h:person)))) ; not in HTML4

(drd-table tmhtml-with-cmd%
  (("font family" "tt") (h:tt))
  (("font series" "bold") (h:b))
  (("font shape" "italic") (h:i)))

(drd-table tmhtml-with-cmd% ; deprecated
  ("font size" ,tmhtml-with-font-size)
  (("color" "black") (h:font (@ ("color" "black"))))
  (("color" "grey") (h:font (@ ("color" "grey"))))
  (("color" "white") (h:font (@ ("color" "white"))))
  (("color" "red") (h:font (@ ("color" "red"))))
  (("color" "blue") (h:font (@ ("color" "blue"))))
  (("color" "yellow") (h:font (@ ("color" "black"))))
  (("color" "magenta") (h:font (@ ("color" "magenta"))))
  (("color" "orange") (h:font (@ ("color" "orange"))))
  (("color" "green") (h:font (@ ("color" "green"))))
  (("color" "brown") (h:font (@ ("color" "brown"))))
  (("paragraph mode" "left") (h:div (@ ("align" "left"))))
  (("paragraph mode" "justify") (h:div (@ ("align" "justify"))))
  (("paragraph mode" "center") (h:center)))

(drd-table tmhtml-with-cmd% ; netscape4
  (("nr columns" "1") (h:multicol (@ ("cols" "1"))))
  (("nr columns" "2") (h:multicol (@ ("cols" "2"))))
  (("nr columns" "3") (h:multicol (@ ("cols" "3"))))
  (("nr columns" "4") (h:multicol (@ ("cols" "4"))))
  (("nr columns" "5") (h:multicol (@ ("cols" "5")))))

(drd-dispatcher tmhtml-tables%
  (block (lambda (l) (tmhtml-table "l" #t l)))
  (block* (lambda (l) (tmhtml-table "c" #t l)))
  (tabular (lambda (l) (tmhtml-table "l" #f l)))
  (tabular* (lambda (l) (tmhtml-table "c" #f l))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Produce patch for keeping commands understood by the converter
;;   FIXME: we should use the DRD here
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-env-macro-0 name)
  `(associate ,(symbol->string name)
	      (macro (hold (,name)))))

(define (tmhtml-env-macro-1 name)
  `(associate ,(symbol->string name)
	      (macro "x" (hold (,name (release (arg "x")))))))

(define (tmhtml-env-patch)
  `(collection
    ,@(map tmhtml-env-macro-0
	   '(item))
    ,@(map tmhtml-env-macro-1
	   '(chapter section subsection subsubsection paragraph subparagraph
	     itemize itemize-minus itemize-dot itemize-arrow
	     enumerate enumerate-numeric enumerate-roman
	     enumerate-Roman enumerate-alpha enumerate-Alpha
	     description description-compact description-dash
	     description-align description-long item*
	     strong em dfn code* samp kbd var abbr acronym tt
	     block block* tabular tabular*))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-produce x)
  ;; Convert a TeXmacs tree x to a HTML tree.
  ;;
  ;; If we are converting a document, the tree x is:
  ;;   (file! <doc> <style> <lan> <tmpath>)
  ;; with:
  ;;   <doc>: edit tree of the document being converted.
  ;;   <style>: TUPLE of style and package names used by the document. If
  ;;     there is only a style (or no style at all) a TUPLE is introduced by
  ;;     the caller.
  ;;   <lan>: the language of the document (from init env).
  ;;   <tmpath>: TEXMACS_PATH with trailing slashes removed.
  ;;
  ;; If we are converting a simple snippet (exporting the clipboard) the tree x
  ;; is simply the snippet being exported.
  (htmlout-produce (sxml-strip-ns-prefix "h" (tmhtml x))))
